/*
 * PICBUT.C
 *
 * Contains the main window procedure of the Picture Button control
 * that handles mouse logic, and Windows messages.
 *
 * Robin J. Leatherbarrow
 * Erithacus Software Ltd.
 *
 * Largely based on the MUSCROLL code by Kraig Brockschmidt
 * found in the SDK
 */


#include <windows.h>
#include "picbdll.h"



/*
 * PicButWndProc
 *
 * Purpose:
 *	Window Procedure for the PictureButton custom control.	Handles all
 *  messages like WM_PAINT just as a normal application window would.
 *  Any message not processed here should go to DefWindowProc.
 *
 * Parameters:
 *  hWnd            HWND handle to the control window.
 *  iMsg            WORD message identifier.
 *  wParam          WORD parameter of the message.
 *  lParam          LONG parameter of the message.
 *
 * Return Value:
 *  LONG            Varies with the message.
 *
 */

LONG FAR __export PASCAL PicButWndProc(HWND hWnd, UINT iMsg,
									   WPARAM wParam, LPARAM lParam)
    {
	PPICBUT			pPB;
	int 			width, height;
	int 			newwidth, newheight;
	POINT			pt;
	UINT			uState;
	RECT			rect;


    /*
	 * Get a pointer to the PICBUT structure for this control.
     * Note that if we do this before WM_NCCREATE where we allocate
	 * the memory, pPB will be NULL, which is not a problem since
     * we do not access it until after WM_NCCREATE.
     */
	pPB=(PPICBUT)GetWindowWord(hWnd, GWW_PICBUTHMEM);


    //Let the API handler process WM_USER+xxxx messages
    if (iMsg >= WM_USER)
		return LPicButAPI(hWnd, iMsg, wParam, lParam, pPB);


    //Handle standard Windows messages.
    switch (iMsg)
        {
        case WM_NCCREATE:
        case WM_CREATE:
			return LPicButCreate(hWnd, iMsg, pPB, (LPCREATESTRUCT)lParam);

        case WM_NCDESTROY:
			//Free the bitmap resource
			if (pPB->hBit)
				if (pPB->bDeleteBitmap)
					DeleteObject( pPB->hBit );

			//Free the control's memory.
			LocalFree((HANDLE)pPB);
            break;

        case WM_ERASEBKGND:
            /*
             * Eat this message to avoid erasing portions that
             * we are going to repaint in WM_PAINT.  Part of a
             * change-state-and-repaint strategy is to rely on
             * WM_PAINT to do anything visual, which includes
             * erasing invalid portions.  Letting WM_ERASEBKGND
             * erase the background is redundant.
             */
            break;

		case WM_SIZE:
			// the bIgnoreSize flag lets us
			// avoid infinite loops...
			//
			if (!pPB->bIgnoreSize) {

				newwidth = width = LOWORD(lParam);	/* width of client area  */
				newheight = height = HIWORD(lParam); /* height of client area */

				if (pPB->bLimitSize) {

					newwidth = max(newwidth, pPB->width + TOOLBUTTON_EXTRAWIDTH);
					newheight = max(newheight, pPB->height + TOOLBUTTON_EXTRAHEIGHT);

					if (!pPB->hBit) {
						newwidth = TOOLBUTTON_STDWIDTH;
						newheight = TOOLBUTTON_STDHEIGHT;
						}

					if ((width != newwidth) || (height != newheight)) {

						pPB->bIgnoreSize = TRUE;

						SetWindowPos(hWnd, NULL, 0, 0,
									 newwidth,
									 newheight,
									 SWP_NOZORDER | SWP_NOMOVE);

						pPB->bIgnoreSize = FALSE;
						}


					}
				}

			break;

		case WM_SETTEXT:
			//Free the bitmap resource
			if (pPB->hBit)
				if (pPB->bDeleteBitmap)
					DeleteObject( pPB->hBit );
			pPB->hBit = NULL;

			ResetImage( hWnd, pPB, (LPSTR)lParam );
			break;


		case WM_PAINT:
			return PicButPaint( hWnd, pPB );

        case WM_ENABLE:
            /*
			 * Handles disabling/enabling case.
			 */

			if (wParam) {
				StateClear(pPB, BUTTONGROUP_DISABLED);
				StateSet(pPB, BUTTONGROUP_ACTIVE);
				}
			else {
				StateClear(pPB, BUTTONGROUP_ACTIVE);
				StateSet(pPB, BUTTONGROUP_DISABLED);
				}

            //Force a repaint since the control will look different.
            InvalidateRect(hWnd, NULL, TRUE);
            UpdateWindow(hWnd);
            break;



        case WM_CANCELMODE:
            /*
             * IMPORTANT MESSAGE!  WM_CANCELMODE means that a
             * dialog or some other modal process has started.
             * we must make sure that we cancel any clicked state
             * we are in, kill the timers, and release the capture.
             */
			if (hWnd==GetCapture()) {
				if (pPB->uState != pPB->uOldState) {
					pPB->uState = pPB->uOldState;
					InvalidateRect(hWnd, NULL, TRUE);
					UpdateWindow(hWnd);
					}
				ReleaseCapture();
				}
            break;



		case WM_LBUTTONDBLCLK:
        case WM_LBUTTONDOWN:
            /*
			 * A WM_LBUTTONDBLCLK message means that the user clicked
             * the control twice in succession which we want to treat
             * like WM_LBUTTONDOWN.  This is safe since we will receive
             * WM_LBUTTONUP before the WM_LBUTTONDBLCLK.
             */

			// nothing if we are disabled
			if (StateTest(pPB, BUTTONGROUP_DISABLED))
				break;


			// save the old state
			pPB->uOldState = pPB->uState;

			// set the state to mouse down
			if (pPB->dwStyle & PBS_COMMAND)
				pPB->uState = COMMANDBUTTON_MOUSEDOWN;
			else
				pPB->uState = ATTRIBUTEBUTTON_MOUSEDOWN;

			InvalidateRect(hWnd, NULL, TRUE);
            UpdateWindow(hWnd);


			// do we inform with a WM_MENUSELECT command?
			if (pPB->dwStyle & PBS_MENUSELECT)
				SendMessage( GetParent(hWnd), WM_MENUSELECT,
							 GetWindowWord(hWnd, GWW_ID),
							 0L );

            SetCapture(hWnd);
            break;


		case WM_MOUSEMOVE:
			// only important if we have captured the mouse
			//
			if (GetCapture() != hWnd)
				break;

			// what is the current style?
			uState = pPB->uState;

			// are we over the control?
			pt=MAKEPOINT(lParam);
			GetClientRect( hWnd, &rect );
			if (PtInRect(&rect, pt)) {
				// set the state to mouse down
				if (pPB->dwStyle & PBS_COMMAND)
					pPB->uState = COMMANDBUTTON_MOUSEDOWN;
				else
					pPB->uState = ATTRIBUTEBUTTON_MOUSEDOWN;
				}
			else {
				pPB->uState = pPB->uOldState;
				}

			// if different then show it
			//
			if (uState!=pPB->uState) {
				InvalidateRect(hWnd, NULL, TRUE);
                UpdateWindow(hWnd);
                }

            break;


		case WM_LBUTTONUP:
			ProcessButtonup( hWnd, lParam, pPB	);
			break;



		default:
			return DefWindowProc(hWnd, iMsg, wParam, lParam);
        }

    return 0L;
    }
